@@ -31,8 +31,8 @@
 #include "gs-window.h"
 #include "gs-job.h"
 #include "gs-grab.h"
-
-#include "fade.h"
+#include "gs-fade.h"
+#include "gs-debug.h"
 
 static void gs_manager_class_init (GSManagerClass *klass);
 static void gs_manager_init       (GSManager      *manager);
@@ -55,8 +55,10 @@ struct GSManagerPrivate
         guint        user_switch_enabled : 1;
         guint        throttle_enabled : 1;
 
+        char        *logout_command;
+
         /* State */
-        guint        blank_active : 1;
+        guint        active : 1;
         guint        lock_active : 1;
 
         guint        dialog_up : 1;
@@ -64,15 +66,15 @@ struct GSManagerPrivate
         guint        lock_timeout_id;
         guint        cycle_timeout_id;
 
-        time_t       blank_time;
+        time_t       activate_time;
 
         GSList      *themes;
         GSSaverMode  saver_mode;
 };
 
 enum {
-        BLANKED,
-        UNBLANKED,
+        ACTIVATED,
+        DEACTIVATED,
         LAST_SIGNAL
 };
 
@@ -84,14 +86,17 @@ enum {
         PROP_LOCK_TIMEOUT,
         PROP_CYCLE_TIMEOUT,
         PROP_LOGOUT_TIMEOUT,
-        PROP_BLANK_ACTIVE,
+        PROP_LOGOUT_COMMAND,
+        PROP_ACTIVE,
         PROP_THROTTLE_ENABLED,
 };
 
+#define FADE_TIMEOUT 1000
+
 static GObjectClass *parent_class = NULL;
 static guint         signals [LAST_SIGNAL] = { 0, };
 
-G_DEFINE_TYPE (GSManager, gs_manager, G_TYPE_OBJECT);
+G_DEFINE_TYPE (GSManager, gs_manager, G_TYPE_OBJECT)
 
 void
 gs_manager_set_mode (GSManager  *manager,
@@ -213,6 +218,24 @@ activate_lock_timeout (GSManager *manager)
         return FALSE;
 }
 
+static void
+remove_lock_timer (GSManager *manager)
+{
+        if (manager->priv->lock_timeout_id != 0) {
+                g_source_remove (manager->priv->lock_timeout_id);
+                manager->priv->lock_timeout_id = 0;
+        }
+}
+
+static void
+add_lock_timer (GSManager *manager,
+                glong      timeout)
+{
+        manager->priv->lock_timeout_id = g_timeout_add (timeout,
+                                                        (GSourceFunc)activate_lock_timeout,
+                                                        manager);
+}
+
 void
 gs_manager_set_lock_timeout (GSManager *manager,
                              glong      lock_timeout)
@@ -223,23 +246,18 @@ gs_manager_set_lock_timeout (GSManager *manager,
 
                 manager->priv->lock_timeout = lock_timeout;
 
-                if (manager->priv->blank_active
-                    && !manager->priv->lock_active
+                if (manager->priv->active
+                    && ! manager->priv->lock_active
                     && (lock_timeout >= 0)) {
 
-                        glong elapsed = (time (NULL) - manager->priv->blank_time) * 1000;
+                        glong elapsed = (time (NULL) - manager->priv->activate_time) * 1000;
 
-                        if (manager->priv->lock_timeout_id) {
-                                g_source_remove (manager->priv->lock_timeout_id);
-                                manager->priv->lock_timeout_id = 0;
-                        }
+                        remove_lock_timer (manager);
 
                         if (elapsed >= lock_timeout) {
                                 activate_lock_timeout (manager);
                         } else {
-                                manager->priv->lock_timeout_id = g_timeout_add (lock_timeout - elapsed,
-                                                                                (GSourceFunc)activate_lock_timeout,
-                                                                                manager);
+                                add_lock_timer (manager, lock_timeout - elapsed);
                         }
                 }
         }
@@ -260,6 +278,21 @@ gs_manager_set_logout_timeout (GSManager *manager,
         }
 }
 
+void
+gs_manager_set_logout_command (GSManager  *manager,
+                               const char *command)
+{
+        g_return_if_fail (GS_IS_MANAGER (manager));
+
+        g_free (manager->priv->logout_command);
+
+        if (command) {
+                manager->priv->logout_command = g_strdup (command);
+        } else {
+                manager->priv->logout_command = NULL;
+        }
+}
+
 static const char *
 select_theme (GSManager *manager)
 {
@@ -272,7 +305,6 @@ select_theme (GSManager *manager)
                 return NULL;
 
         if (manager->priv->saver_mode == GS_MODE_DONT_BLANK) {
-                g_warning ("Selecting no screensaver theme since screensaver is disabled.");
                 return NULL;
         }
 
@@ -299,7 +331,7 @@ gs_manager_cycle (GSManager *manager)
         g_return_val_if_fail (manager != NULL, FALSE);
         g_return_val_if_fail (GS_IS_MANAGER (manager), FALSE);
 
-        if (! manager->priv->blank_active)
+        if (! manager->priv->active)
                 return FALSE;
 
         if (manager->priv->dialog_up)
@@ -334,6 +366,24 @@ cycle_timeout (GSManager *manager)
         return TRUE;
 }
 
+static void
+remove_cycle_timer (GSManager *manager)
+{
+        if (manager->priv->cycle_timeout_id != 0) {
+                g_source_remove (manager->priv->cycle_timeout_id);
+                manager->priv->cycle_timeout_id = 0;
+        }
+}
+
+static void
+add_cycle_timer (GSManager *manager,
+                 glong      timeout)
+{
+        manager->priv->cycle_timeout_id = g_timeout_add (timeout,
+                                                         (GSourceFunc)cycle_timeout,
+                                                         manager);
+}
+
 void
 gs_manager_set_cycle_timeout (GSManager *manager,
                               glong      cycle_timeout)
@@ -344,23 +394,20 @@ gs_manager_set_cycle_timeout (GSManager *manager,
 
                 manager->priv->cycle_timeout = cycle_timeout;
 
-                if (manager->priv->blank_active && (cycle_timeout >= 0)) {
+                if (manager->priv->active && (cycle_timeout >= 0)) {
                         glong timeout;
-                        glong elapsed = (time (NULL) - manager->priv->blank_time) * 1000;
+                        glong elapsed = (time (NULL) - manager->priv->activate_time) * 1000;
 
-                        if (manager->priv->cycle_timeout_id) {
-                                g_source_remove (manager->priv->cycle_timeout_id);
-                                manager->priv->cycle_timeout_id = 0;
-                        }
+                        remove_cycle_timer (manager);
 
                         if (elapsed >= cycle_timeout) {
                                 timeout = 0;
                         } else {
                                 timeout = cycle_timeout - elapsed;
                         }
-                        manager->priv->cycle_timeout_id = g_timeout_add (timeout,
-                                                                         (GSourceFunc)cycle_timeout,
-                                                                         manager);
+
+                        add_cycle_timer (manager, timeout);
+
                 }
         }
 }
@@ -394,6 +441,9 @@ gs_manager_set_property (GObject            *object,
         case PROP_LOGOUT_TIMEOUT:
                 gs_manager_set_logout_timeout (self, g_value_get_long (value));
                 break;
+        case PROP_LOGOUT_COMMAND:
+                gs_manager_set_logout_command (self, g_value_get_string (value));
+                break;
         case PROP_CYCLE_TIMEOUT:
                 gs_manager_set_cycle_timeout (self, g_value_get_long (value));
                 break;
@@ -432,11 +482,14 @@ gs_manager_get_property (GObject            *object,
         case PROP_LOGOUT_TIMEOUT:
                 g_value_set_long (value, self->priv->logout_timeout);
                 break;
+        case PROP_LOGOUT_COMMAND:
+                g_value_set_string (value, self->priv->logout_command);
+                break;
         case PROP_CYCLE_TIMEOUT:
                 g_value_set_long (value, self->priv->cycle_timeout);
                 break;
-        case PROP_BLANK_ACTIVE:
-                g_value_set_boolean (value, self->priv->blank_active);
+        case PROP_ACTIVE:
+                g_value_set_boolean (value, self->priv->active);
         default:
                 G_OBJECT_WARN_INVALID_PROPERTY_ID (object, prop_id, pspec);
                 break;
@@ -454,21 +507,21 @@ gs_manager_class_init (GSManagerClass *klass)
         object_class->get_property = gs_manager_get_property;
         object_class->set_property = gs_manager_set_property;
 
-        signals [BLANKED] =
-                g_signal_new ("blanked",
+        signals [ACTIVATED] =
+                g_signal_new ("activated",
                               G_TYPE_FROM_CLASS (object_class),
                               G_SIGNAL_RUN_LAST,
-                              G_STRUCT_OFFSET (GSManagerClass, blanked),
+                              G_STRUCT_OFFSET (GSManagerClass, activated),
                               NULL,
                               NULL,
                               g_cclosure_marshal_VOID__VOID,
                               G_TYPE_NONE,
                               0);
-        signals [UNBLANKED] =
-                g_signal_new ("unblanked",
+        signals [DEACTIVATED] =
+                g_signal_new ("deactivated",
                               G_TYPE_FROM_CLASS (object_class),
                               G_SIGNAL_RUN_LAST,
-                              G_STRUCT_OFFSET (GSManagerClass, unblanked),
+                              G_STRUCT_OFFSET (GSManagerClass, deactivated),
                               NULL,
                               NULL,
                               g_cclosure_marshal_VOID__VOID,
@@ -476,8 +529,8 @@ gs_manager_class_init (GSManagerClass *klass)
                               0);
 
         g_object_class_install_property (object_class,
-                                         PROP_BLANK_ACTIVE,
-                                         g_param_spec_boolean ("blank-active",
+                                         PROP_ACTIVE,
+                                         g_param_spec_boolean ("active",
                                                                NULL,
                                                                NULL,
                                                                FALSE,
@@ -522,6 +575,13 @@ gs_manager_class_init (GSManagerClass *klass)
                                                             G_MAXLONG,
                                                             0,
                                                             G_PARAM_READWRITE));
+        g_object_class_install_property (object_class,
+                                         PROP_LOGOUT_TIMEOUT,
+                                         g_param_spec_string ("logout-command",
+                                                              NULL,
+                                                              NULL,
+                                                              NULL,
+                                                              G_PARAM_READWRITE));
         g_object_class_install_property (object_class,
                                          PROP_CYCLE_TIMEOUT,
                                          g_param_spec_long ("cycle-timeout",
@@ -549,18 +609,10 @@ gs_manager_init (GSManager *manager)
 }
 
 static void
-remove_blank_timers (GSManager *manager)
+remove_timers (GSManager *manager)
 {
-
-        if (manager->priv->lock_timeout_id) {
-                g_source_remove (manager->priv->lock_timeout_id);
-                manager->priv->lock_timeout_id = 0;
-        }
-
-        if (manager->priv->cycle_timeout_id) {
-                g_source_remove (manager->priv->cycle_timeout_id);
-                manager->priv->cycle_timeout_id = 0;
-        }
+        remove_lock_timer (manager);
+        remove_cycle_timer (manager);
 }
 
 static void
@@ -576,7 +628,9 @@ gs_manager_finalize (GObject *object)
 
         g_return_if_fail (manager->priv != NULL);
 
-        remove_blank_timers (manager);
+        g_free (manager->priv->logout_command);
+
+        remove_timers (manager);
 
         gs_grab_release_keyboard_and_mouse ();
 
@@ -593,32 +647,34 @@ gs_manager_finalize (GObject *object)
         g_slist_free (manager->priv->windows);
         manager->priv->windows = NULL;
 
-        manager->priv->blank_active = FALSE;
-        manager->priv->blank_time = 0;
+        manager->priv->active = FALSE;
+        manager->priv->activate_time = 0;
         manager->priv->lock_enabled = FALSE;
 
         G_OBJECT_CLASS (parent_class)->finalize (object);
 }
 
 static gboolean
-window_unblank_idle (GSManager *manager)
+window_deactivated_idle (GSManager *manager)
 {
         g_return_val_if_fail (manager != NULL, FALSE);
         g_return_val_if_fail (GS_IS_MANAGER (manager), FALSE);
-        
-        gs_manager_unblank (manager);
+
+        /* don't deactivate directly but only emit a signal
+           so that we let the parent deactivate */
+        g_signal_emit (manager, signals [DEACTIVATED], 0);
 
         return FALSE;
 }
 
 static void
-window_unblanked_cb (GSWindow  *window,
-                     GSManager *manager)
+window_deactivated_cb (GSWindow  *window,
+                       GSManager *manager)
 {
         g_return_if_fail (manager != NULL);
         g_return_if_fail (GS_IS_MANAGER (manager));
         
-        g_idle_add ((GSourceFunc)window_unblank_idle, manager);
+        g_idle_add ((GSourceFunc)window_deactivated_idle, manager);
 }
 
 static void
@@ -725,31 +781,59 @@ window_show_cb (GSWindow  *window,
 
         manager->priv->jobs = g_slist_append (manager->priv->jobs, job);
 
-        manager->priv->blank_time = time (NULL);
+        manager->priv->activate_time = time (NULL);
 
         /* Set lock state to off initially.  It may be enabled by activate_lock_timout() */
         gs_manager_set_lock_active (manager, FALSE);
 
         if (manager->priv->lock_timeout >= 0) {
-                if (manager->priv->lock_timeout_id)
-                        g_source_remove (manager->priv->lock_timeout_id);
-
-                manager->priv->lock_timeout_id = g_timeout_add (manager->priv->lock_timeout,
-                                                                (GSourceFunc)activate_lock_timeout,
-                                                                manager);
+                remove_lock_timer (manager);
+                add_lock_timer (manager, manager->priv->lock_timeout);
         }
 
         if (manager->priv->cycle_timeout >= 10000) {
-                if (manager->priv->cycle_timeout_id)
-                        g_source_remove (manager->priv->cycle_timeout_id);
-
-                manager->priv->cycle_timeout_id = g_timeout_add (manager->priv->cycle_timeout,
-                                                                 (GSourceFunc)cycle_timeout,
-                                                                 manager);
+                remove_cycle_timer (manager);
+                add_cycle_timer (manager, manager->priv->cycle_timeout);
         }
 
         /* FIXME: only emit signal once */
-        g_signal_emit (manager, signals [BLANKED], 0);
+        g_signal_emit (manager, signals [ACTIVATED], 0);
+}
+
+static void
+disconnect_window_signals (GSManager *manager,
+                           GSWindow  *window)
+{
+        g_signal_handlers_disconnect_by_func (window, window_deactivated_cb, manager);
+        g_signal_handlers_disconnect_by_func (window, window_dialog_up_cb, manager);
+        g_signal_handlers_disconnect_by_func (window, window_dialog_down_cb, manager);
+        g_signal_handlers_disconnect_by_func (window, window_show_cb, manager);
+        g_signal_handlers_disconnect_by_func (window, window_map_event_cb, manager);
+}
+
+static void
+window_destroyed_cb (GtkWindow *window,
+                     GSManager *manager)
+{
+        disconnect_window_signals (manager, GS_WINDOW (window));
+}
+
+static void
+connect_window_signals (GSManager *manager,
+                        GSWindow  *window)
+{
+        g_signal_connect_object (window, "destroy",
+                                 G_CALLBACK (window_destroyed_cb), manager, 0);
+        g_signal_connect_object (window, "deactivated",
+                                 G_CALLBACK (window_deactivated_cb), manager, 0);
+        g_signal_connect_object (window, "dialog-up",
+                                 G_CALLBACK (window_dialog_up_cb), manager, 0);
+        g_signal_connect_object (window, "dialog-down",
+                                 G_CALLBACK (window_dialog_down_cb), manager, 0);
+        g_signal_connect_object (window, "show",
+                                 G_CALLBACK (window_show_cb), manager, 0);
+        g_signal_connect_object (window, "map_event",
+                                 G_CALLBACK (window_map_event_cb), manager, 0);
 }
 
 static void
@@ -772,20 +856,12 @@ gs_manager_create_window (GSManager *manager,
         for (i = 0; i < n_monitors; i++) {
                 window = gs_window_new (screen, i, manager->priv->lock_active);
 
-                gs_window_set_logout_enabled (window, manager->priv->logout_enabled);
                 gs_window_set_user_switch_enabled (window, manager->priv->user_switch_enabled);
+                gs_window_set_logout_enabled (window, manager->priv->logout_enabled);
                 gs_window_set_logout_timeout (window, manager->priv->logout_timeout);
+                gs_window_set_logout_command (window, manager->priv->logout_command);
 
-                g_signal_connect_object (window, "unblanked",
-                                         G_CALLBACK (window_unblanked_cb), manager, 0);
-                g_signal_connect_object (window, "dialog-up",
-                                         G_CALLBACK (window_dialog_up_cb), manager, 0);
-                g_signal_connect_object (window, "dialog-down",
-                                         G_CALLBACK (window_dialog_down_cb), manager, 0);
-                g_signal_connect_object (window, "show",
-                                         G_CALLBACK (window_show_cb), manager, 0);
-                g_signal_connect_object (window, "map_event",
-                                         G_CALLBACK (window_map_event_cb), manager, 0);
+                connect_window_signals (manager, window);
 
                 manager->priv->windows = g_slist_append (manager->priv->windows, window);
         }
@@ -827,49 +903,78 @@ gs_manager_new (void)
         return GS_MANAGER (manager);
 }
 
-gboolean
-gs_manager_blank (GSManager *manager)
+static void
+show_windows (GSList *windows)
+{
+        GSList *l;
+
+        for (l = windows; l; l = l->next) {
+                gtk_widget_show (GTK_WIDGET (l->data));
+        }
+}
+
+static gboolean
+gs_manager_activate (GSManager *manager)
 {
         GdkDisplay *display;
         GdkScreen  *screen;
         GdkWindow  *root;
+        gboolean    do_fade;
 
         g_return_val_if_fail (manager != NULL, FALSE);
         g_return_val_if_fail (GS_IS_MANAGER (manager), FALSE);
 
-        if (manager->priv->blank_active)
+        if (manager->priv->active) {
+                g_warning ("Trying to activate manager when already active");
                 return FALSE;
-
-        manager->priv->blank_active = TRUE;
+        }
 
         display = gdk_display_get_default ();
         gdk_display_get_pointer (display, &screen, NULL, NULL, NULL);
         root = gdk_screen_get_root_window (screen);
 
-        if (! gs_grab_get_keyboard_and_mouse (root, screen))
+        if (! gs_grab_get_keyboard_and_mouse (root, screen)) {
                 return FALSE;
+        }
 
         if (! manager->priv->windows)
                 gs_manager_create (GS_MANAGER (manager));
 
         /* fade to black and show windows */
-        fade_screens_and_show (3, 20, TRUE, manager->priv->windows);
+        do_fade = TRUE;
+        if (do_fade) {
+                GSFade *fade;
+
+                fade = gs_fade_new ();
+                gs_fade_set_timeout (fade, FADE_TIMEOUT);
+                gs_fade_now (fade);
+                show_windows (manager->priv->windows);
+                g_usleep (500000);
+                gs_fade_reset (fade);
+                g_object_unref (fade);
+        } else {
+                show_windows (manager->priv->windows);
+        }
+
+        manager->priv->active = TRUE;
 
         return TRUE;
 }
 
-gboolean
-gs_manager_unblank (GSManager *manager)
+static gboolean
+gs_manager_deactivate (GSManager *manager)
 {
         GSList *l;
 
         g_return_val_if_fail (manager != NULL, FALSE);
         g_return_val_if_fail (GS_IS_MANAGER (manager), FALSE);
 
-        if (! manager->priv->blank_active)
+        if (! manager->priv->active) {
+                g_warning ("Trying to deactivate a screensaver that is not active");
                 return FALSE;
+        }
 
-        remove_blank_timers (manager);
+        remove_timers (manager);
 
         gs_grab_release_keyboard_and_mouse ();
 
@@ -886,23 +991,35 @@ gs_manager_unblank (GSManager *manager)
         g_slist_free (manager->priv->windows);
         manager->priv->windows = NULL;
 
-        manager->priv->blank_active = FALSE;
-        manager->priv->blank_time = 0;
+        manager->priv->active = FALSE;
+        manager->priv->activate_time = 0;
         manager->priv->lock_active = FALSE;
         manager->priv->throttle_enabled = FALSE;
 
-        g_signal_emit (manager, signals [UNBLANKED], 0);
-
         return TRUE;
 }
 
 gboolean
-gs_manager_is_blanked (GSManager *manager)
+gs_manager_set_active (GSManager *manager,
+                       gboolean   active)
+{
+        gboolean res;
+
+        if (active)
+                res = gs_manager_activate (manager);
+        else
+                res = gs_manager_deactivate (manager);
+
+        return res;
+}
+
+gboolean
+gs_manager_is_active (GSManager *manager)
 {
         g_return_val_if_fail (manager != NULL, FALSE);
         g_return_val_if_fail (GS_IS_MANAGER (manager), FALSE);
 
-        return manager->priv->blank_active;
+        return manager->priv->active;
 }
 
 void
@@ -916,7 +1033,7 @@ gs_manager_request_unlock (GSManager *manager)
         g_return_if_fail (manager != NULL);
         g_return_if_fail (GS_IS_MANAGER (manager));
 
-        if (! manager->priv->blank_active)
+        if (! manager->priv->active)
                 return;
 
         /* Find the screen that contains the pointer */
