@@ -32,11 +32,13 @@
 #include "gnome-screensaver.h"
 
 #include "gs-manager.h"
-#include "gs-watcher-x11.h"
+#include "gs-watcher.h"
+#include "gs-fade.h"
 #include "gs-power.h"
 #include "gs-listener-dbus.h"
 #include "gs-monitor.h"
 #include "gs-prefs.h"
+#include "gs-debug.h"
 
 static void     gs_monitor_class_init (GSMonitorClass *klass);
 static void     gs_monitor_init       (GSMonitor      *monitor);
@@ -51,15 +53,18 @@ struct GSMonitorPrivate
         GSManager      *manager;
         GSPower        *power;
         GSPrefs        *prefs;
+        GSFade         *fade;
 };
 
 enum {
         PROP_0
 };
 
+#define FADE_TIMEOUT 10000
+
 static GObjectClass *parent_class = NULL;
 
-G_DEFINE_TYPE (GSMonitor, gs_monitor, G_TYPE_OBJECT);
+G_DEFINE_TYPE (GSMonitor, gs_monitor, G_TYPE_OBJECT)
 
 static void
 gs_monitor_class_init (GSMonitorClass *klass)
@@ -74,14 +79,14 @@ gs_monitor_class_init (GSMonitorClass *klass)
 }
 
 static void
-manager_blanked_cb (GSManager *manager,
-                    GSMonitor *monitor)
+manager_activated_cb (GSManager *manager,
+                      GSMonitor *monitor)
 {
 }
 
 static void
-manager_unblanked_cb (GSManager *manager,
-                      GSMonitor *monitor)
+manager_deactivated_cb (GSManager *manager,
+                         GSMonitor *monitor)
 {
         gs_listener_set_active (monitor->priv->listener, FALSE);
 }
@@ -92,18 +97,52 @@ watcher_idle_cb (GSWatcher *watcher,
                  GSMonitor *monitor)
 {
         gboolean res;
+        gboolean fade_active;
+
+        fade_active = gs_fade_get_active (monitor->priv->fade);
 
         res = gs_listener_set_idle (monitor->priv->listener, TRUE);
 
+        if (fade_active) {
+                gs_fade_set_active (monitor->priv->fade, FALSE);
+        }
+
         return res;
 }
 
+static gboolean
+watcher_idle_notice_cb (GSWatcher *watcher,
+                        int        reserved,
+                        GSMonitor *monitor)
+{
+        /* start slow fade */
+        gs_fade_set_timeout (monitor->priv->fade, FADE_TIMEOUT);
+        gs_fade_set_active (monitor->priv->fade, TRUE);
+
+        return TRUE;
+}
+
+static void
+watcher_notice_cancelled_cb (GSWatcher *watcher,
+                             GSMonitor *monitor)
+{
+        /* cancel the fade */
+        gs_fade_set_active (monitor->priv->fade, FALSE);
+        gs_fade_reset (monitor->priv->fade);
+}
+
 static void
 listener_lock_cb (GSListener *listener,
                   GSMonitor  *monitor)
 {
-        gs_listener_set_active (monitor->priv->listener, TRUE);
-        gs_manager_set_lock_active (monitor->priv->manager, TRUE);
+        gboolean res;
+
+        res = gs_listener_set_active (monitor->priv->listener, TRUE);
+        if (res) {
+                gs_manager_set_lock_active (monitor->priv->manager, TRUE);
+        } else {
+                g_warning ("Unable to lock the screen");
+        }
 }
 
 static void
@@ -121,30 +160,73 @@ listener_cycle_cb (GSListener *listener,
         gs_manager_cycle (monitor->priv->manager);
 }
 
-static void
+static gboolean
 listener_active_changed_cb (GSListener *listener,
                             gboolean    active,
                             GSMonitor  *monitor)
 {
+        gboolean res;
+        gboolean idle_watch_enabled;
+
+        idle_watch_enabled = gs_watcher_get_enabled (monitor->priv->watcher);
+
         if (active) {
-                /* turn off the idleness watcher */
-                gs_watcher_set_active (monitor->priv->watcher, FALSE);
+                if (idle_watch_enabled) {
+                        /* turn off the idleness watcher */
+                        res = gs_watcher_set_active (monitor->priv->watcher, FALSE);
+                        if (! res) {
+                                g_warning ("Unable to deactivate the idle watcher");
+                                return FALSE;
+                        }
+                }
 
                 /* blank the screen */
-                gs_manager_blank (monitor->priv->manager);
+                res = gs_manager_set_active (monitor->priv->manager, TRUE);
+                if (! res) {
+                        g_warning ("Unable to blank the screen");
+
+                        /* since we can't activate then reactivate the watcher
+                           and give up */
+                        if (idle_watch_enabled) {
+                                res = gs_watcher_set_active (monitor->priv->watcher, TRUE);
+                        }
+
+                        return FALSE;
+                }
 
                 /* enable power management */
-                gs_power_set_active (monitor->priv->power, TRUE);
+                res = gs_power_set_active (monitor->priv->power, TRUE);
+                if (! res) {
+                        g_warning ("Unable to activate power management");
+
+                        /* if we can't activate power management it isn't the
+                           end of the world */
+                }
         } else {
                 /* unblank the screen */
-                gs_manager_unblank (monitor->priv->manager);
+                res = gs_manager_set_active (monitor->priv->manager, FALSE);
+                if (! res) {
+                        g_warning ("Unable to unblank the screen");
+                        return FALSE;
+                }
 
-                /* turn on the idleness watcher */
-                gs_watcher_set_active (monitor->priv->watcher, TRUE);
+                if (idle_watch_enabled) {
+                        /* turn on the idleness watcher */
+                        res = gs_watcher_set_active (monitor->priv->watcher, TRUE);
+                        if (! res) {
+                                g_warning ("Unable to activate the idle watcher");
+                                return FALSE;
+                        }
+                }
 
                 /* disable power management */
-                gs_power_set_active (monitor->priv->power, FALSE);
+                res = gs_power_set_active (monitor->priv->power, FALSE);
+                if (! res) {
+                        g_warning ("Unable to deactivate power management");
+                }
         }
+
+        return TRUE;
 }
 
 static void
@@ -195,19 +277,36 @@ static void
 _gs_monitor_update_from_prefs (GSMonitor *monitor,
                                GSPrefs   *prefs)
 {
-        gs_manager_set_lock_enabled (monitor->priv->manager, monitor->priv->prefs->lock);
+        gboolean idle_detection_enabled;
+        gboolean idle_detection_active;
+        gboolean activate_watch;
+        gboolean manager_active;
+
+        gs_manager_set_lock_enabled (monitor->priv->manager, monitor->priv->prefs->lock_enabled);
         gs_manager_set_lock_timeout (monitor->priv->manager, monitor->priv->prefs->lock_timeout);
         gs_manager_set_logout_enabled (monitor->priv->manager, monitor->priv->prefs->logout_enabled);
         gs_manager_set_user_switch_enabled (monitor->priv->manager, monitor->priv->prefs->user_switch_enabled);
         gs_manager_set_logout_timeout (monitor->priv->manager, monitor->priv->prefs->logout_timeout);
+        gs_manager_set_logout_command (monitor->priv->manager, monitor->priv->prefs->logout_command);
         gs_manager_set_cycle_timeout (monitor->priv->manager, monitor->priv->prefs->cycle);
         gs_manager_set_mode (monitor->priv->manager, monitor->priv->prefs->mode);
         gs_manager_set_themes (monitor->priv->manager, monitor->priv->prefs->themes);
 
-        gs_watcher_set_active (monitor->priv->watcher,
-                               monitor->priv->prefs->mode != GS_MODE_DONT_BLANK);
-
+        idle_detection_enabled = (monitor->priv->prefs->mode != GS_MODE_DONT_BLANK);
         gs_watcher_set_timeout (monitor->priv->watcher, monitor->priv->prefs->timeout);
+        gs_watcher_set_enabled (monitor->priv->watcher, idle_detection_enabled);
+
+        /* in the case where idle detection is reenabled we may need to
+           activate the watcher too */
+
+        manager_active = gs_manager_is_active (monitor->priv->manager);
+        idle_detection_active = gs_watcher_get_active (monitor->priv->watcher);
+        activate_watch = (! manager_active
+                          && ! idle_detection_active
+                          && idle_detection_enabled);
+        if (activate_watch) {
+                gs_watcher_set_active (monitor->priv->watcher, TRUE);
+        }
 
         gs_power_set_timeouts (monitor->priv->power,
                                monitor->priv->prefs->dpms_standby,
@@ -218,16 +317,19 @@ _gs_monitor_update_from_prefs (GSMonitor *monitor,
 }
 
 static void
-gs_monitor_init (GSMonitor *monitor)
+disconnect_listener_signals (GSMonitor *monitor)
 {
+        g_signal_handlers_disconnect_by_func (monitor->priv->listener, listener_lock_cb, monitor);
+        g_signal_handlers_disconnect_by_func (monitor->priv->listener, listener_quit_cb, monitor);
+        g_signal_handlers_disconnect_by_func (monitor->priv->listener, listener_cycle_cb, monitor);
+        g_signal_handlers_disconnect_by_func (monitor->priv->listener, listener_active_changed_cb, monitor);
+        g_signal_handlers_disconnect_by_func (monitor->priv->listener, listener_throttled_changed_cb, monitor);
+        g_signal_handlers_disconnect_by_func (monitor->priv->listener, listener_poke_cb, monitor);
+}
 
-        monitor->priv = GS_MONITOR_GET_PRIVATE (monitor);
-
-        monitor->priv->prefs = gs_prefs_new ();
-        g_signal_connect_swapped (monitor->priv->prefs, "changed",
-                                  G_CALLBACK (_gs_monitor_update_from_prefs), monitor);
-
-        monitor->priv->listener = gs_listener_new ();
+static void
+connect_listener_signals (GSMonitor *monitor)
+{
         g_signal_connect (monitor->priv->listener, "lock",
                           G_CALLBACK (listener_lock_cb), monitor);
         g_signal_connect (monitor->priv->listener, "quit",
@@ -240,23 +342,92 @@ gs_monitor_init (GSMonitor *monitor)
                           G_CALLBACK (listener_throttled_changed_cb), monitor);
         g_signal_connect (monitor->priv->listener, "poke",
                           G_CALLBACK (listener_poke_cb), monitor);
+}
 
-        monitor->priv->watcher = gs_watcher_new (monitor->priv->prefs->timeout);
+static void
+disconnect_watcher_signals (GSMonitor *monitor)
+{
+        g_signal_handlers_disconnect_by_func (monitor->priv->watcher, watcher_idle_cb, monitor);
+        g_signal_handlers_disconnect_by_func (monitor->priv->watcher, watcher_idle_notice_cb, monitor);
+        g_signal_handlers_disconnect_by_func (monitor->priv->watcher, watcher_notice_cancelled_cb, monitor);
+}
+
+static void
+connect_watcher_signals (GSMonitor *monitor)
+{
         g_signal_connect (monitor->priv->watcher, "idle",
                           G_CALLBACK (watcher_idle_cb), monitor);
+        g_signal_connect (monitor->priv->watcher, "idle_notice",
+                          G_CALLBACK (watcher_idle_notice_cb), monitor);
+        g_signal_connect (monitor->priv->watcher, "notice_cancelled",
+                          G_CALLBACK (watcher_notice_cancelled_cb), monitor);
+}
 
-        monitor->priv->manager = gs_manager_new ();
-        g_signal_connect (monitor->priv->manager, "blanked",
-                          G_CALLBACK (manager_blanked_cb), monitor);
-        g_signal_connect (monitor->priv->manager, "unblanked",
-                          G_CALLBACK (manager_unblanked_cb), monitor);
+static void
+disconnect_manager_signals (GSMonitor *monitor)
+{
+        g_signal_handlers_disconnect_by_func (monitor->priv->manager, manager_activated_cb, monitor);
+        g_signal_handlers_disconnect_by_func (monitor->priv->manager, manager_deactivated_cb, monitor);
+}
 
-        monitor->priv->power = gs_power_new ();
+static void
+connect_manager_signals (GSMonitor *monitor)
+{
+        g_signal_connect (monitor->priv->manager, "activated",
+                          G_CALLBACK (manager_activated_cb), monitor);
+        g_signal_connect (monitor->priv->manager, "deactivated",
+                          G_CALLBACK (manager_deactivated_cb), monitor);
+}
+
+static void
+disconnect_power_signals (GSMonitor *monitor)
+{
+        g_signal_handlers_disconnect_by_func (monitor->priv->power, power_changed_cb, monitor);
+}
+
+static void
+connect_power_signals (GSMonitor *monitor)
+{
         g_signal_connect (monitor->priv->power, "changed",
                           G_CALLBACK (power_changed_cb), monitor);
+}
+
+static void
+disconnect_prefs_signals (GSMonitor *monitor)
+{
+        g_signal_handlers_disconnect_by_func (monitor->priv->prefs, _gs_monitor_update_from_prefs, monitor);
+}
+
+static void
+connect_prefs_signals (GSMonitor *monitor)
+{
+        g_signal_connect_swapped (monitor->priv->prefs, "changed",
+                                  G_CALLBACK (_gs_monitor_update_from_prefs), monitor);
+}
+
+static void
+gs_monitor_init (GSMonitor *monitor)
+{
+
+        monitor->priv = GS_MONITOR_GET_PRIVATE (monitor);
+
+        monitor->priv->prefs = gs_prefs_new ();
+        connect_prefs_signals (monitor);
+
+        monitor->priv->listener = gs_listener_new ();
+        connect_listener_signals (monitor);
+
+        monitor->priv->fade = gs_fade_new ();
+        monitor->priv->watcher = gs_watcher_new (monitor->priv->prefs->timeout);
+        connect_watcher_signals (monitor);
+
+        monitor->priv->manager = gs_manager_new ();
+        connect_manager_signals (monitor);
+
+        monitor->priv->power = gs_power_new ();
+        connect_power_signals (monitor);
 
         _gs_monitor_update_from_prefs (monitor, monitor->priv->prefs);
-
 }
 
 static void
@@ -271,9 +442,17 @@ gs_monitor_finalize (GObject *object)
 
         g_return_if_fail (monitor->priv != NULL);
 
+        disconnect_watcher_signals (monitor);
+        disconnect_listener_signals (monitor);
+        disconnect_manager_signals (monitor);
+        disconnect_power_signals (monitor);
+        disconnect_prefs_signals (monitor);
+
+        g_object_unref (monitor->priv->fade);
         g_object_unref (monitor->priv->watcher);
         g_object_unref (monitor->priv->listener);
         g_object_unref (monitor->priv->manager);
+        g_object_unref (monitor->priv->power);
         g_object_unref (monitor->priv->prefs);
 
         G_OBJECT_CLASS (parent_class)->finalize (object);
