@@ -33,6 +33,7 @@
 
 #include "gs-window.h"
 #include "subprocs.h"
+#include "gs-debug.h"
 
 static void gs_window_class_init (GSWindowClass *klass);
 static void gs_window_init       (GSWindow      *window);
@@ -323,6 +324,63 @@ add_emit_deactivated_idle (GSWindow *window)
         g_idle_add ((GSourceFunc)emit_deactivated_idle, window);
 }
 
+static void
+gs_window_raise (GSWindow *window)
+{
+        GdkWindow *win;
+
+        g_return_if_fail (GS_IS_WINDOW (window));
+
+        win = GTK_WIDGET (window)->window;
+
+        gdk_window_raise (win);
+}
+
+static void
+gs_window_xevent (GSWindow  *window,
+                  GdkXEvent *xevent)
+{
+        XEvent *ev;
+
+        ev = xevent;
+
+        /* MapNotify is used to tell us when new windows are mapped.
+           ConfigureNofify is used to tell us when windows are raised. */
+        switch (ev->xany.type) {
+        case MapNotify:
+                gs_window_raise (window);
+                break;
+        case ConfigureNotify:
+                gs_window_raise (window);
+                break;
+        default:
+                break;
+        }
+
+}
+
+static GdkFilterReturn
+xevent_filter (GdkXEvent *xevent,
+               GdkEvent  *event,
+               GSWindow  *window)
+{
+        gs_window_xevent (window, xevent);
+
+        return GDK_FILTER_CONTINUE;
+}
+
+static void
+select_popup_events (void)
+{
+        XWindowAttributes attr;
+        unsigned long     events;
+
+        XGetWindowAttributes (GDK_DISPLAY (), GDK_ROOT_WINDOW (), &attr);
+
+        events = SubstructureNotifyMask | attr.your_event_mask;
+        XSelectInput (GDK_DISPLAY (), GDK_ROOT_WINDOW(), events);
+}
+
 static void
 gs_window_real_show (GtkWidget *widget)
 {
@@ -340,6 +398,9 @@ gs_window_real_show (GtkWidget *widget)
 
         remove_watchdog_timer (window);
         add_watchdog_timer (window, 30000);
+
+        select_popup_events ();
+        gdk_window_add_filter (NULL, (GdkFilterFunc)xevent_filter, window);
 }
 
 void
@@ -357,6 +418,8 @@ gs_window_real_hide (GtkWidget *widget)
 
         window = GS_WINDOW (widget);
 
+        gdk_window_remove_filter (NULL, (GdkFilterFunc)xevent_filter, window);
+
         remove_watchdog_timer (window);
 
         if (GTK_WIDGET_CLASS (parent_class)->hide)
